% SVM with missing data algorith with fast search algorithm
% create on 07/02/2008
% update on 07/06/2008, add  isNormalizeU, K_opt, U_opt
function [w,b] = PSVM(X, M, y, isNormalizeU, K_opt, U_opt)
% X is the data after filling in
% M is the mask of visibility
% y is the label of classes (-1, +1)
% isNormalizeU to determine the normalization of U
% K_opt is the parameter to control weights K
% U_opt is the parameter to control U (or u)

if (nargin < 3)||(nargin > 6)
    w = Inf;    b = Inf;    
elseif (nargin == 3)
    isNormalizeU = true; K_opt = 0; U_opt = 0;
elseif (nargin == 4)
    K_opt = 0; U_opt = 0;
elseif (nargin == 5)
    U_opt = 0;
end;
d = size(X,1);
% K is the weight of each data with missing components
switch K_opt
   case {0} % K increases as miss# increases
      K = single(d) - single(sum(M)); 
   case {1} % K set to 1 if missing data exists
      K = 1*((single(d) - single(sum(M))) > 0);
   case {2} % K decreases as miss# increases
      K = single(sum(M)); 
   otherwise
      K = single(d) - single(sum(M)); 
end


% figure();hold on;
% for i = 1 : size(X,2) 
%     if y(i)==-1 plot(X(1,i),X(2,i),'r.');
%     else if y(i)==1  plot(X(1,i),X(2,i),'b.'); end; end;
% end;
% hold off;

% construct Y
Y = zeros(size(X,2),d+1);
for i = 1 : size(X,2)
    Y(i,:) = [y(i)*X(:,i)',-y(i)];
end;
I = eye(d); u = 1.0*zeros(1,d);
for j = 1 : d  u(j) = sum(K.*M(j,:)); end;
% normalize u
if isNormalizeU
    u = u / norm(u);
end;
% 
switch U_opt
   case {0} % keep u
      u = 1*u; 
   case {1} % double u
      u = 2*u;
   case {2}
      u = u.^2; 
   case {3}
      u = 10*u; 
   case {4}
      u = 0.1*u; 
   case {5}
      u = 20*u; 
   case {6}
      u = 40*u; 
   case {7}
      u = 0.05*u; 
   case {8}
      u = 0.025*u; 
   otherwise
      u = 1*u; 
end




U = diag(u); 
r0 = max(u);

counter_cvx = 0;
% the first run of CVX
r = r0;
tic;
step = 1;
step = r0/100;
isFindInitial = false;
while ~isFindInitial
    r_init = r + step;
    cvx_begin
        variable w(d);
        variable b;
        minimize(quad_form(w,r_init*I-U));
        subject to
            Y*([w;b]) >= 1;
            quad_form(w,r_init*I-U) <= 1;
        cvx_quiet(true);
    cvx_end
    counter_cvx = counter_cvx+1;
    if strcmp(cvx_status,'Solved')
        display('Find initial value!');
        isFindInitial = true;
        r = r_init;
        disp(['r_init = ',num2str(r_init,'%.6f')]);
    else
        step = step/2;
    end;
end;

results_final = cvx_optpnt;
r_new = (1+w'*U*w)/(w'*w);
r_diff = r_new - r;
r = r_new;
disp(['r = ',num2str(r,'%.6f')]);

% if strcmp(cvx_status,'Solved')
%     display('Initialized value passed!');
% else
%     display('Initialized value failed!');
%     return;
% end;

while r_diff>1e-4%1e-6
    cvx_begin
        variable w(d);
        variable b;
        minimize(quad_form(w,r*I-U));
        subject to
            Y*([w;b]) >= 1;
            quad_form(w,r*I-U) <= 1;
        cvx_quiet(true);
    cvx_end
    counter_cvx = counter_cvx+1;
    if strcmp(cvx_status,'Solved')
        % step_diff = 1 - cvx_optval;
        results_final = cvx_optpnt;
        r_new = (1+w'*U*w)/(w'*w);
        r_diff = r_new - r;
        r = r_new;
        disp(['r = ',num2str(r,'%.6f')]);
    else
        r_diff = 0;
    end;
end;
t=toc
disp(['CVX runs ',num2str(counter_cvx,'%d'),' times by fast search!']);

w = results_final.w;
b = results_final.b;
